pi: float
e: float
tau: float
inf: float
nan: float
def sqrt(__x: float) -> float: ...
def sin(__x: float) -> float: ...
def cos(__x: float) -> float: ...
def tan(__x: float) -> float: ...
def exp(__x: float) -> float: ...
def log(__x: float) -> float: ...
def floor(__x: float) -> int: ...
def ceil(__x: float) -> int: ...
def fabs(__x: float) -> float: ...
def pow(__x: float, __y: float) -> float: ...
def copysign(__x: float, __y: float) -> float: ...
def isinf(__x: float) -> bool: ...
def isnan(__x: float) -> bool: ...
def isfinite(__x: float) -> bool: ...
def nextafter(__x: float, __y: float) -> float: ...
