from mypy_extensions import u8

class Buffer:
    def __init__(self, source: bytes = ...) -> None: ...
    def getvalue(self) -> bytes: ...

def write_bool(data: Buffer, value: bool) -> None: ...
def read_bool(data: Buffer) -> bool: ...
def write_str(data: Buffer, value: str) -> None: ...
def read_str(data: Buffer) -> str: ...
def write_float(data: Buffer, value: float) -> None: ...
def read_float(data: Buffer) -> float: ...
def write_int(data: Buffer, value: int) -> None: ...
def read_int(data: Buffer) -> int: ...
def write_tag(data: Buffer, value: u8) -> None: ...
def read_tag(data: Buffer) -> u8: ...
