import sys
from typing_extensions import deprecated

if sys.version_info >= (3, 14):
    @deprecated("The `nturl2path` module is deprecated since Python 3.14.")
    def url2pathname(url: str) -> str: ...
    @deprecated("The `nturl2path` module is deprecated since Python 3.14.")
    def pathname2url(p: str) -> str: ...

else:
    def url2pathname(url: str) -> str: ...
    def pathname2url(p: str) -> str: ...
